# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions to transform files via a search loop.  This is meant to be used to
implemnent the filter scripts, and provided as infrastructure for the user to
write his own filtering scripts.
"""

__author__ = "Martin Blais <blais@furius.ca>"


# stdlib imports.
import sys, os, tempfile
from os.path import *

# xxdiff imports.
import xxdiff.scripts
import xxdiff.selectfiles
import xxdiff.backup
import xxdiff.invoke
import xxdiff.checkout
import xxdiff.condrepl
from xxdiff.scripts import tmpprefix


__all__ = ['Transformer', 'parse_args', 'transform_replace_loop']


class Transformer(object):
    """
    Base class for all transformers.
    """
    def __init__(self, opts):
        """
        'opts': an options object -> Option

          We're expecting to find opts.dry_run on it.
        """
        self.opts = opts

    def transform(self, fn, outf):
        """
        Process the file 'fn' and write the transformed contents to the
        temporary file 'tmpf' (this file object is already opened in 'w' mode.

        Return True if the file if the transformation was succesful, False if
        the file should be skipped.
        """
        raise NotImplementedError


def parse_args(parser):
    """
    Parse the options and return:
    - an options object -> Options
    - program arguments -> list of strings
    - a selector of the files to iterate over -> generator
    """
    xxmodules = (xxdiff.selectfiles, xxdiff.backup, xxdiff.checkout,
                 xxdiff.invoke, xxdiff.condrepl)
    for mod in xxmodules:
        mod.options_graft(parser)

    xxdiff.scripts.install_autocomplete(parser)

    # Parse arguments
    opts, args = parser.parse_args()

    vargs = opts, parser, sys.stdout
    for mod in xxmodules[1:]:
        mod.options_validate(*vargs)
    selector = xxdiff.selectfiles.options_validate(*vargs)

    return opts, args, selector


def transform_replace_loop(opts, selector, xformer, logs):
    """
    Run a loop through the selected files and conditionally transform them by
    applying the given Transformer and invoking xxdiff to confirm.  Return a map
    of the filenames to the decision code that was used on each of them.
    """
    decision_map = {}

    for fn in selector:
        # Create temporary file to receive the transformed file.
        tmpf = tempfile.NamedTemporaryFile(mode='w', prefix=tmpprefix)

        # Transform the file.
        if xformer.transform(fn, tmpf) is False:
            # This file is to be skipped by the transformer for some reason.
            decision_map[fn] = 'SKIPPED'
            continue

        # Flush the temporary file before invoking xxdiff on it.
        tmpf.flush()

        # Invoke conditional replacement via xxdiff.
        decision = xxdiff.condrepl.cond_replace(fn, tmpf.name, opts, logs)
        decision_map[fn] = decision

    return decision_map


def postloop_footer(opts, decision_map=None):
    """
    Print some common stuff at the end of the looping scripts.
    You don't have to add this to your script, but it's convenient.
    """
    xxdiff.backup.print_reminder(opts)




# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
xx-find-grep-sed [<options>] <regexp> <sed-cmd>

Perform global sed-like replacements in a set of files.

This program walks a directory hierarchy, selects some files to be processed by
filenames (or by default, process all files found), then each selected file is
searched for some regexp pattern, and if there is a match, run the given file
through a sed command and replace the file with that output.

This script effectively becomes obsolete, because it can be replace by a much
more generic equivalent::

   xx-filter --select-grep='<regexp>' 'sed -e <sed-code>'

"""



__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4', 'diffutils', 'sed']

# stdlib imports.
import sys, os, re
from os.path import *
from subprocess import Popen, PIPE

# xxdiff imports.
import xxdiff.scripts
import xxdiff.xformloop


class GrepSedTransformer(xxdiff.xformloop.Transformer):
    """
    Transformer that greps the file for a regular expression, and if it matches,
    runs it through a sed command.
    """
    def __init__(self, opts, regexp, sedprog):
        xxdiff.xformloop.Transformer.__init__(self, opts)
        self.regexp, self.sedprog = regexp, sedprog

    def transform(self, fn, outf):
        # Grep the input file.
        try:
            text = open(fn, 'r').read()
        except IOError as e:
            # Don't bail out if we cannot read a file.
            print("Warning: cannot read file '%s'." % fn, file=sys.stderr)
            return False

        # Grep the file.
        if not self.regexp.search(text, re.MULTILINE):
            return False

        # If this is not for show only, make sure right away that we can
        # overwrite the destination file
        if self.opts.dry_run and not os.access(fn, os.W_OK):
            raise SystemExit("Error: cannot write to file '%s'." % fn)

        # Perform sed replacement to a temporary file
        p = Popen(['sed', '-e', self.sedprog],
                  stdin=PIPE, stdout=outf, stderr=PIPE, text=True)
        try:
            stdout, stderr = p.communicate(text)
        except OSError as e:
            raise SystemExit("Error: running sed command '%s':\n%s" %
                             (self.sedprog, p.stderr.read()))

        if p.returncode != 0:
            raise SystemExit(
                "Error: running sed command:\nFile:%s\n%s" % (fn, stderr))

        return True


def parse_options():
    import optparse
    parser = optparse.OptionParser(__doc__.strip())
    opts, args, selector = xxdiff.xformloop.parse_args(parser)

    # Check that we got two arguments
    if len(args) < 2:
        parser.error("You must specify at least a regular expression and "
                     "a sed command")
    regexp, sedprog = args

    # Compile the regular expression
    try:
        regexp = re.compile(regexp)
    except re.error:
        parser.error("Cannot compile given regexp '%s'." % regexp)

    # Force to always perform a diff on output.
    opts.verbose = 2

    return regexp, sedprog, opts, selector


def findgrepsed_main():
    """
    Main program for find-grep-sed script.
    """
    # Parse the options.
    regexp, sedprog, opts, selector = parse_options()

    # Create an appropriate transformer.
    xformer = GrepSedTransformer(opts, regexp, sedprog)

    import warnings
    warnings.warn("xx-find-grep-sed is obsolete. "
                  "Use xx-filter with --select-grep option instead.")

    try:
        decision_map = xxdiff.xformloop.transform_replace_loop(
            opts, selector, xformer, sys.stdout)
    finally:
        xxdiff.xformloop.postloop_footer(opts)


def main():
    xxdiff.scripts.interruptible_main(findgrepsed_main)

if __name__ == '__main__':
    main()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""xx-diff-proxy [<options>] <file> <file> [<file>]

Simple wrapper script for xxdiff that presents options that are the same as the
diff or diff3 commands.  This can be used with programs that expect to call
diff/diff3 to present differences to users, and that can be configured with a
different diff program.

This was originally created for integration with Subversion.  This is meant to
be used in a user's configuration file to set diff-cmd and diff3-cmd options, so
that when you run ``svn diff`` it spawns xxdiffs instead of textual diffs (see
also xx-svn-diff).

Add the following lines ``~/.subversion/config``::

  [helpers]
  diff-cmd = xx-diff-proxy
  diff3-cmd = xx-diff-proxy

This script returns the same value as the underlying xxdiff with option
--exit-with-merge-status, so that you can do something special if there are
remaining conflicts.  I think Subversion uses that on 'svn update' if you
configure it in your helper hooks.
"""

__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4']


# stdlib imports.
import sys, os

# xxdiff imports.
import xxdiff.scripts
import xxdiff.invoke


def parse_options():
    """
    Parse the options.
    """
    import optparse
    parser = optparse.OptionParser(__doc__.strip())

    xxdiff.invoke.options_graft(parser)

    parser.add_option('--decision', action='store_true',
                      help="Use decision mode for the diffed files.  This "
                      "allows you to save merged files over the originals.")


    # For diff-cmd invocation.
    parser.add_option('-u', action='store_true',
                      help="See diff(1). Ignored for xxdiff.")

    parser.add_option('-L', '--label', action='append', default=[],
                      help="See diff(1). Label gets set as file title string "
                      "for xxdiff.")

    # For diff3-cmd invocation.
    parser.add_option('-E', '--show-overlap', action='store_true',
                      help="See diff3(1). Ignored for xxdiff.")

    parser.add_option('-m', '--merge', action='store_true',
                      help="See diff3(1). Passed to xxdiff for automatic "
                      "selection of non-conflictual regions..")

    ## parser.add_option('--reverse', action='store_true',
    ##                   help="Reverse the display order of the files.")

    xxdiff.scripts.install_autocomplete(parser)
    opts, args = parser.parse_args()

    if len(args) > 3:
        raise parser.error("Cannot invoke wrapper with more than 3 files.")

    xxdiff.invoke.options_validate(opts, parser)

    return opts, args


def diffproxy_main():
    """
    Main program for cond-replace script.
    """
    opts, args = parse_options()

    dargs = ['--exit-with-merge-status']
    if opts.merge:
        dargs.append('--merge')

    for o1, o2, fn in zip(['--title1', '--title2', '--title3'],
                          ['"%s"' % x for x in opts.label],
                          args):
        if os.path.basename(fn) == 'empty-file':
            o2 = '"(NON-EXISTANT FILE)"'

        dargs.extend([o1, o2])

    dargs.extend(args)

    # Decision xxdiff: if we force it or if we have 3 arguments.
    if opts.decision or len(args) == 3:
        decision, mergedf, retcode = xxdiff.invoke.xxdiff_decision(opts, *dargs)

        # If the user merged, copy the merged file over the original.
        if decision in ('MERGED', 'ACCEPT'):
            mergedf.flush()
            mergedf.seek(0)
            text = mergedf.read()
            mergedf.close()
            sys.stdout.write(text)
    
        elif decision in ('NODECISION', 'REJECT'):
            pass # do nothing
        else:
            raise SystemExit(
                    "Error: unexpected answer from xxdiff: %s" % decision)

    # Just spawn xxdiff and discard the results.
    else:
        retcode = xxdiff.invoke.xxdiff_display(opts, *dargs)
        
    # Process return code
    if retcode not in [0, 3]:
        raise SystemExit('Warning: Unexpected xxdiff exit code: %d' % retcode)

    return retcode


main = diffproxy_main

if __name__ == '__main__':
    main()





#!/usr/bin/env python3
# This file is part of the xxdiff package.  See xxdiff for license and details.

# If this file is found independently as svn-foreign, here are the licensing
# terms:
#
#  svn-foreign
#  Copyright (C) 2006  Martin Blais
#  URL:  http://furius.ca/xxdiff/
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""svn-foreign [<options>] [<dir> ...]

This script allows you to quickly add or delete files that are foreign to a
Subversion checkout, and to easily resolve conflicts.  It deals with the common
case where you have developed something using Subversion and you have forgotten
some unregistered files before committing.

svn-foreign runs 'svn status' on the given Subversion-managed directories, to
find out which files are unaccounted for and for each of these files, it asks
you what to do with it::

- add it
- delete it
- ignore/mask it (sets an svn:ignore property on its parent directory)
- skip it for now (leave it where it is)

Other actions::

- quit / [x] exit, this interrupts the process
- delete with no backups (for large files)
- view the file with a pager (more)
- resolve conflict (only valid on a file with an unresolved conflict)
- revert conflict (only valid on a file with an unresolved conflict)

The script works interactively and is meant to allow you to quickly deal with
the forgotten files in a subversion checkout.  It works with directories as
well.

Also, files in conflicts are properly ignored and not counted as foreign files.
When unresolved conflicts are found, you are queried about them, answer with 'r'
to resolve it and delete the temporary files.

"""

# Future Work
# -----------
#
# - Make backups work without the xxdiff.backup module.
# - svn-foreign: add support for editing conflictual file
#
# Credits
# -------
# * To Sean Reifschneider (Jafo) for providing example code for raw tty input



__version__ = '$Revision$'
__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports
import sys, os, termios, tty, tempfile, datetime, re
from subprocess import Popen, PIPE, call
from os.path import *

# xxdiff imports (optional)
try:
    from xxdiff import backup
except ImportError:
    backup = None


debug = False

def read_one():
    """
    Reads a single character from the terminal.
    """
    # Set terminal in raw mode for faster input.
    orig_term_attribs = termios.tcgetattr(sys.stdin.fileno())
    tty.setraw(sys.stdin.fileno())
    try:
        c = sys.stdin.read(1)
        if c == chr(3): # INTR
            raise KeyboardInterrupt
    finally:
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSAFLUSH,
                          orig_term_attribs)
    return c

def add(fn):
    """
    Add the file into Subversion.
    """
    call(['svn', 'add', fn])

def delete(fn):
    """
    Delete the file from Subversion.
    """
    call(['svn', 'delete', fn])

def rmrf(fnodn):
    """
    Delete the given directory and all its contents.
    """
    if not exists(fnodn):
        return

    if isdir(fnodn):
        for root, dirs, files in os.walk(fnodn, topdown=False):
            for fn in files:
                os.remove(join(root, fn))
            for dn in dirs:
                adn = join(root, dn)
                if islink(adn):
                    os.remove(adn)
                else:
                    os.rmdir(adn)
        os.rmdir(root)
    else:
        os.remove(fnodn)

def ignore_prop(dn, ignores):
    """
    Set svn:ignore on directory 'dn'.
    """
    f = tempfile.NamedTemporaryFile(mode='w')
    f.write(ignores)
    f.flush()
    call(['svn', 'propset', 'svn:ignore', '-F', f.name, dn])
    f.close()


def filter2(predicate, *arguments):
    ein, eout = [], []
    for args in zip(*arguments):
        add_args = args
        if len(args) == 1:
            add_args = args[0]

        if predicate(*args):
            ein.append(add_args)
        else:
            eout.append(add_args)
    return ein, eout


def parse_options():
    """
    Parse the options.
    """
    import optparse
    parser = optparse.OptionParser(__doc__.strip())

    parser.add_option('-c', '--commit', action='store_true',
                      help="Ask to commit after running.")

    parser.add_option('-m', '--comments', action='store',
                      help="Specify comments for commit.")

    parser.add_option('-v', '--verbose',
                      action='store_const', const=2, dest='verbose',
                      default=1,
                      help=optparse.SUPPRESS_HELP)
    parser.add_option('-q', '--quiet', '--silent',
                      action='store_const', const=0, dest='verbose',
                      help="Suppress certain harmless warnings.")

    # Add backups disabling option if the xxdiff libraries are available.
    if backup is not None:
        backup.options_graft(parser,
                             "These options affect automatic backup of "
                             "deleted files.")

    opts, args = parser.parse_args()

    if backup is not None:
        backup.options_validate(opts, parser)

        if opts.backup_prefix is None:
            # Set a better default for the backup prefix.
            opts.backup_prefix = (
                datetime.datetime.now().replace(microsecond=0).isoformat('T'))

    if opts.comments and not opts.commit:
        parser.error("You cannot specify comments if you're not going to "
                     "commit!")

    return opts, args


def view(fn, write):
    "Call on 'more' to view the file."
    write('-' * 80 + '\n')
    pager = os.environ.get('PAGER', '/bin/more')
    call([pager, fn])


def query_unregistered_svn_files(filenames, opts, output=sys.stdout,
                                 ignore=[]):
    """
    Runs an 'svn status' command, and then loops over all the files that are not
    registered, asking the user one-by-one what action to take (see this
    module's docstring for more details).

    'ignore' specifies a list of files to completely ignore.

    Return True upon completion.  Anything else signals that the user has
    interrupted the procedure.
    """
    write = output.write

    if not hasattr(opts, 'verbose'):
        # svnforeign requires a verbose option.
        raise RuntimeError(
            "Internal Error: You are missing a 'verbose' option.")

    ignore = list(map(abspath, ignore))

    # If there are multiple args--such as *--filter out the non svn directories.
    dirs, files = filter2(isdir, filenames)
    if len(dirs) > 1:
        svndirs = [x for x in dirs if isdir(join(x, '.svn'))]
        if svndirs != dirs:
            if opts.verbose >= 1:
                for dn in set(dirs) - set(svndirs):
                    write("Warning: skipping non-checkout dir '%s'\n" % dn)
            filenames = svndirs + files

    # Run svn status to find the files that are unaccounted for
    cmd = ['svn', 'status'] + filenames
    if debug:
        write(' '.join(cmd))
    p = Popen(cmd, stdout=PIPE, stderr=PIPE, text=True)
    out, err = p.communicate()
    if p.returncode != 0:
        raise SystemExit("Error: %s" % err)

    # Get out if there is nothing to do.
    if not out.strip():
        write('(Nothing to do.)\n')
        return True

    # Print the output of the status command for the user's excitement
    write('Status\n------\n')
    write(out)
    write('\n')

    # Figure out which files are in conflict so that we can ignore the unknown
    # merge result files.
    conflicts = []
    for line in out.splitlines():
        if not line or re.match('^Performing', line):
            continue

        if line[0] == 'C':
            fn = line[8:]
            cre = re.compile('%s\\.(mine|r\\d+)' % re.escape(fn)).match
            conflicts.append( (fn, cre) )

            # Command loop (one command)
            while True:
                write('=> [Resolve|Skip|View|Quit|Redo]  %s ? ' % fn)

                # Read command
                c = read_one()
                write(c)
                write('\n')

                if c == 'r': # Add
                    call(['svn', 'resolved', fn])
                    break

                elif c == 's': # Skip
                    break

                elif c in 'v': # View
                    view(fn, write)

                elif c in ['q', 'x']: # Quit/exit
                    write('(Quitting.)\n')
                    return False

                elif c == 'r': # Restart from scratch
                    # (perhaps you invoke that after adding an ignore wildcard.)
                    return query_unregistered_svn_files(filenames, opts, output, ignore)

    # Process foreign files
    for line in out.splitlines():
        if not line or re.match('^(Performing|Status)', line):
            continue

        if line[0] in '?~':
            fn = line[8:]
            if [x for x in conflicts if x[1](fn)]:
                continue
            if abspath(fn) in ignore:
                continue

            # Get the file size.
            try:
                size = os.lstat(fn).st_size
            except OSError as e:
                if e.errno == 2:
                    continue # File not found, was probably a temp file.
                else:
                    raise

            ftype = ''
            if islink(fn):
                ftype = '(symlink)'

            # Command loop (one command)
            while True:
                write('=> [Add|Delete|Ignore|Skip|View|Quit|Redo]  %10d  %s %s ? ' %
                      (size, fn, ftype))

                # Read command
                c = read_one()
                write(c)
                write('\n')

                if c == 'a': # Add
                    add(fn)
                    break

                elif c == 'd': # Delete
                    # Do backups if requested.
                    if backup is not None:
                        backupfn = backup.backup_file(fn, opts, output)
                        if backupfn:
                            write("Backed up to: '%s'\n" % backupfn)
                    elif opts.verbose >= 1:
                        write('(Warning: no backups.)\n')
                    rmrf(fn)
                    break

                elif c == 'D': # Delete with no backups
                    rmrf(fn)
                    break

                elif c in ['m', 'i', 'I']: # Ignore (Mask, svn:ignore)
                    dn, bn = split(fn)
                    if dn == '':
                        dn = '.'

                    # Get the properties first
                    p = Popen(['svn', 'propget', 'svn:ignore', dn],
                              stdout=PIPE, stderr=PIPE, text=True)
                    svnign, err = p.communicate()
                    if p.returncode != 0:
                        raise SystemExit(err)

                    # Sanitize the svn:ignore contents a bit
                    if svnign:
                        svnign = svnign.rstrip() + '\n'

                    write("--(current svn:ignore on '%s')--\n" % dn)
                    write(svnign)
                    write('----------------------------\n')

                    write("Add pattern (! to cancel, * to edit, default: [%s]):"
                          % bn)
                    pat = input()
                    if pat == '':
                        pat = bn

                    if pat == '!':
                        print('(cancelled.)')

                    elif pat == '*':
                        call(['svn', 'propedit', 'svn:ignore', dn])
                        break

                    else:
                        svnign += pat + '\n'
                        ignore_prop(dn, svnign)

                        write('----------------------------\n')
                        write(svnign)
                        write('----------------------------\n')

                        break

                elif c == 's': # Skip
                    break

                elif c in ['q', 'x']: # Quit/exit
                    write('(Quitting.)\n')
                    return False

                elif c == 'r': # Restart from scratch
                    # (perhaps you invoke that after adding an ignore wildcard.)
                    return query_unregistered_svn_files(filenames, opts, output, ignore)

                elif c in 'v': # View
                    view(fn, write)

                elif c == chr(12): # Ctrl-L: Clear
                    # FIXME: is there a way to do this directly on the terminal?
                    call(['clear'])

                else: # Loop again
                    write('(Invalid answer.)\n')

        elif line[0] == '!':
            fn = line[8:]
            if abspath(fn) in ignore:
                continue

            # Command loop (one command)
            while True:
                write('=> [Delete|Skip|Quit]  %10d  %s %s ? ' %
                      (-1, fn, ''))

                # Read command
                c = read_one()
                write(c)
                write('\n')

                if c == 'd': # Delete
                    delete(abspath(fn))
                    break

                elif c == 's': # Skip
                    break

                elif c in ['q', 'x']: # Quit/exit
                    write('(Quitting.)\n')
                    return False

                elif c == chr(12): # Ctrl-L: Clear
                    # FIXME: is there a way to do this directly on the terminal?
                    call(['clear'])

                else: # Loop again
                    write('(Invalid answer.)\n')



    write('(Done.)\n')
    return True


def main():
    """
    Main program.
    """
    opts, args = parse_options()

    # Run the main loop...
    if query_unregistered_svn_files(args, opts, sys.stdout) is not True:
        # The user has quit.
        sys.exit(1)

    # Commit the files if requested.
    if opts.commit is not None:
        # Command loop (one command)
        sys.stdout.write('Commit? [Yes|No/Quit] ')

        # Read command
        c = read_one().lower()
        print(c)

        if c == 'y': # Commit.
            comments = opts.comments or ''
            call(['svn', 'commit', '-m', comments] + args)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        print()
        sys.exit(1)



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions to help creating files to compare key/value pairs/triples using
xxdiff.  The code in this module generates files in a way that will diff well,
lining up corresponding items.  It does this by inserting unique but matching
garbage patterns between the key/values.

This is used, for example, to compare objects in SQL schemas from databases.
"""

__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
import os, optparse, tempfile, random, string
from os.path import *

# xxdiff imports.
from xxdiff.scripts import tmpprefix


def options_graft(parser):
    """
    Graft options on given parser for diff maps.
    """
    group = optparse.OptionGroup(parser, "Diff Maps Options",
                                 "These options affect diff maps generation.")

    group.add_option('-g', '--garbage-lines', action='store', type='int',
                     metavar='NBLINES', default=3,
                     help="Specifies the number of garbage lines to use to "
                     "separate the map items.")

    group.add_option('--show-common-only', '--common', action='store_true',
                     help="Only render the keys that are in common in all "
                     "the maps.")

    parser.add_option_group(group)

    return group


def options_validate(opts, parser, logs=None):
    """
    Validate backup options.
    """


def render_diffable_maps(opts, *maps):
    """
    Render the given maps in respective files in a way that they will diff
    nicely to be able to compare corresponding keys side-by-side.  We return the
    temporary files that were created in the process (we return the same number
    of files as the number of given maps).

    Note: random markers may be inserted to improve the quality of the diff.
    """
    # Produce two files that will diff nicely.
    tmpfiles = [tempfile.NamedTemporaryFile(mode='w', prefix=tmpprefix)
                for x in maps]
    
    mapkeys = set()
    for x in maps: mapkeys.update(list(x.keys()))
    mapkeys = sorted(mapkeys)
    for key in mapkeys:
        # Check if some of the keys are missing.  If so, skip this entry.
        if opts.show_common_only:
            if False in [key in x for x in maps]:
                continue

        # Generate random garbage that will match up the entries when diffed.
        ranlines = (['=== %s\n' % str(key)] + 
                    ['===[%s] \n' % ''.join(random.choice(string.letters)
                                            for x in range(70))
                     for x in range(opts.garbage_lines)])

        # Write the output to the files.
        for dmap, tmpf in zip(maps, tmpfiles):
            tmpf.write('\n')
            for line in ranlines:
                tmpf.write(line)
            tmpf.write('\n')
            try:
                tmpf.write(dmap[key])
                tmpf.write('\n\n')
            except KeyError:
                pass

    # Make sure all the files are flushed
    [f.flush() for f in tmpfiles]

    return tmpfiles




# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions for conditional replacement.
"""



__author__ = "Martin Blais <blais@furius.ca>"


# stdlib imports.
import sys, os, optparse, logging
from os.path import *
from subprocess import Popen, PIPE
import shutil

# xxdiff imports.
import xxdiff.backup
import xxdiff.invoke
import xxdiff.checkout


__all__ = ('cond_replace',)


def options_graft(parser):
    """
    Graft options on given parser for automatic file backups.
    """
    group = optparse.OptionGroup(parser, "Replacement options",
                                 "These options affect the overwriting of the"
                                 "original file.")

    group.add_option('-n', '--dry-run', action='store_true',
                     help="Print the commands that would be executed " +
                     "but don't really run them.")

    group.add_option('-x', '--no-confirm', action='store_true',
                     help="Do not ask for confirmation with graphical "
                     "diff viewer.")

    parser.add_option_group(group)

    return group


def options_validate(opts, parser, logs=None):
    """
    Validate replacement options.
    """
    pass


# Side-by-side diff2 command.
sbs_diff_cmd = ['diff', '--side-by-side', '--suppress-common-lines']

def cond_replace(origfn, newfn, opts, logs, exitonsame=False, replfn=None):
    """
    Given two filenames, spawn xxdiff for a decision, and conditionally replace
    the original file with the contents of the new file.  This supports options
    for backups and SCM checkout.

    Arguments:

    - 'origfn': filename to process -> string

    - 'newfn': the new filename to compare -> string

    - 'replfn': if specified, the target filename to replace.

    - 'opts': program options object -> Options instance

      The options that are used are:

      * opts.verbose: How much output to print

          Three levels of verbosity are assumed:
          - 0: completely silent
          - 1: outputs only the decision code and then the filename
               (This can be useful for scripts to process the output of a run)
          - 2: outputs a side-by-side diff of the changes.

      * opts.dry_run: Whether to actually apply the changes or not.

      * opts.no_confirm: Apply the changes without confirmation.

    - 'exitonsame': do not even pop xxdiff if the files are the same

    Returns the decision code for the file.
    """
    origfn = normpath(abspath(origfn))

    # Print header
    if opts.verbose >= 2:
        print('=' * 80, file=logs)
        print('File: ', origfn, file=logs)

    if opts.verbose >= 2 or exitonsame:

        # Run diff between the original and the new/modified file
        p = Popen(sbs_diff_cmd + [origfn, newfn], stdout=PIPE, stderr=PIPE, encoding='utf-8')
        diff_output, stderr = p.communicate()

        # Print differences.
        if p.returncode == 0:
            if opts.verbose >= 2:
                print("(Warning: no differences.)", file=logs)

            if exitonsame:
                print_decision('NODIFF', origfn, opts, logs)
                return 'NODECISION'
    else:
        diff_output = None

    if replfn is None:
        replfn = origfn

    if opts.no_confirm:
        # No graphical diff, just replace the files without asking.
        do_replace_file(replfn, newfn, opts, logs)
        decision = 'NOCONFIRM'

        print_decision(decision, origfn, opts, logs)
        if opts.verbose >= 2 and diff_output: print_diffs(diff_output, logs)
    else:
        # Call xxdiff!
        decision, mergedf, retcode = xxdiff.invoke.xxdiff_decision(
            opts, '--title2', 'NEW FILE', origfn, newfn)

        print_decision(decision, origfn, opts, logs)

        if decision == 'ACCEPT':
            # Changes accepted.
            if replfn != newfn:
                do_replace_file(replfn, newfn, opts, logs)
            if opts.verbose >= 2 and diff_output: print_diffs(diff_output, logs)

        elif decision == 'REJECT':
            # Rejected change (or program killed).
            if replfn != origfn:
                do_replace_file(replfn, origfn, opts, logs)
            if opts.verbose >= 2 and diff_output: print_diffs(diff_output, logs)

        elif decision == 'NODECISION':
            # Program killed, do nothing.
            pass

        elif decision == 'MERGED':
            if opts.verbose >= 2:
                # Run diff again to show the real changes that will be applied.
                p = Popen(sbs_diff_cmd + [origfn, mergedf.name],
                          stdout=PIPE, stderr=PIPE, encoding='utf-8')
                diff_output, stderr = p.communicate()

                if diff_output: print_diffs(diff_output, logs)

            do_replace_file(replfn, mergedf.name, opts, logs)

    return decision


def print_decision(decision, origfn, opts, logs):
    """
    Print the decision string.
    """
    if opts.verbose >= 2:
        print(decision, file=logs)
    elif opts.verbose >= 1:
        print('%-10s %s' % (decision, origfn), file=logs)

def print_diffs(diff_output, logs):
    """
    ppFormat nicely and print the output of side-by-side diff.
    """
    print(file=logs)
    for line in map(str.expandtabs, diff_output.splitlines()):
        print(' |', line, file=logs)
    print(file=logs)

def do_replace_file(ofn, nfn, opts, logs):
    """
    Function that performs the file replacement safely.  Replace the original
    file 'ofn' by 'nfn'.
    """
    if opts.dry_run:
        return

    # Backup the original file first.
    if hasattr(opts, 'backup_type'):
        xxdiff.backup.backup_file(ofn, opts, logs)

    # Insure that the file is checked out.
    if getattr(opts, 'checkout', None):
        xxdiff.checkout.insure_checkout(ofn, opts, logs)

    # Copy the new file over the original.
    if not os.access(ofn, os.W_OK):
        logging.warn("Could not overwrite file '%s'." % ofn)
    else:
        shutil.copyfile(nfn, ofn)



diff3_cmd = ['diff3']

proper_decisions = ('ACCEPT', 'REJECT', 'MERGED')

def cond_resolve(mine, ancestor, yours, output, opts, logs=None, extra=None):
    """
    Given three filenames, 'mine', 'ancestor', 'yours', spawn a 3-way xxdiff for
    a decision, and replace the 'original' file with the contents of the merged
    output.

    Arguments:

    - 'mine', 'ancestor', 'yours': files that contains the local file before
      merging, the common ancestor file of 'mine' and 'yours', and 'yours' the
      merged file, -> strings

    - 'output': the filename where we want to store the merged results (in the
      case of a merge, the original filename) -> string

    - 'opts': program options object -> Options instance

      The options that are used are:

      * opts.verbose: How much output to print

          Three levels of verbosity are assumed:
          - 0: completely silent
          - 1: outputs only the decision code and then the filename
               (This can be useful for scripts to process the output of a run)
          - 2: outputs a side-by-side diff of the changes.

      * opts.dry_run: Whether to actually apply the changes or not.

    - 'extra': are extra parameters to pass on to xxdiff.

    Returns the decision code for the file.
    """
    mine, ancestor, yours, output = list(map(abspath,
                                        (mine, ancestor, yours, output)))
    files3 = [mine, ancestor, yours]

    # Print header
    if opts.verbose >= 2:
        print('=' * 80, file=logs)
        print('File: ', output, file=logs)

    if opts.verbose >= 2:
        # Run diff between the three files.
        p = Popen(diff3_cmd + files3, stdout=PIPE, stderr=PIPE, encoding='utf-8')
        diff_output, stderr = p.communicate()

        # Print differences.
        if not diff_output: # Note: we cannot rely on the return code.
            if opts.verbose >= 2:
                print("(Warning: no differences.)", file=logs)

            print_decision('NODIFF', mine, opts, logs)
            return 'NODECISION'
    else:
        diff_output = None

    # Add arguments to identify files in the title bars.
    dargs = xxdiff.invoke.title_opts('%s (WORKING)' % mine,
                                     '%s (ANCESTOR)' % ancestor,
                                     '%s (MERGED/NEW BASE)' % yours)
    if extra is not None:
        dargs = list(extra) + dargs

    # Call xxdiff!
    dargs.extend(files3)
    decision, mergedf, retcode = xxdiff.invoke.xxdiff_decision(opts, *dargs)

    print_decision(decision, mine, opts, logs)

    # Backup all 3 files and the destination output before overwriting.
    if decision != 'NODECISION' and not opts.dry_run:
        if hasattr(opts, 'backup_type'):
            xxdiff.backup.backup_file(output, opts, logs)

    if decision == 'ACCEPT':
        # Accept: discard our local changes and replace file with the update.
        shutil.copyfile(yours, output)

    elif decision == 'REJECT':
        # Reject: discard the update and keep our local file as it were before
        # the merge.
        shutil.copyfile(mine, output)

    elif decision == 'MERGED':
        # Merge: replace the output file with the merged output of xxdiff.
        shutil.copyfile(mergedf.name, output)

    elif decision == 'NODECISION':
        # No Decision: Do not change anything.
        pass

    if opts.verbose >= 2 and decision in proper_decisions:
        # Run diff again to show the real changes that will be applied to
        # 'mine' into the output file.
        p = Popen(sbs_diff_cmd + [mine, output], stdout=PIPE, stderr=PIPE, encoding='utf-8')
        diff_output, stderr = p.communicate()

        if diff_output:
            print_diffs(diff_output, logs)

    return decision


def test():
    """
    Simple interactive test.
    """
    import tempfile
    from pprint import pprint

    newdir = tempfile.mkdtemp()
    fm = dict([('file%d' % x, join(newdir, 'file%d' % x)) for x in range(1, 5)])
    pprint(fm)
    os.system('cp /home/blais/.emacs %(file1)s' % fm) # TEST
    os.system('cat %(file1)s | sed -e "s/contr/FREAK/" > %(file2)s' % fm) # TEST

    class Opts:
        "dummy options class"
        verbose = 2
        dry_run = False
        no_confirm = False

    print(cond_replace(fm['file1'], fm['file2'], Opts, sys.stdout))
    pprint(fm)



if __name__ == '__main__':
    test()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""xx-svn-diff [<options>] <file> <file> [<file>]

Script that invokes xxdiff for all modified files in the given Subversion
checkout area. Optionally replace the output file by decision mode.
"""



__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4', 'Subversion']

# stdlib imports.
import sys, os, tempfile, datetime
from os.path import *

# xxdiff imports.
import xxdiff.scripts
import xxdiff.invoke
from xxdiff.invoke import title_opts
import xxdiff.editor
import xxdiff.resilient
import xxdiff.utils
from xxdiff.utils import makedirs
import xxdiff.history
import xxdiff.condrepl
from xxdiff.scripts import tmpprefix, script_name
from xxdiff.scm import subversion
from xxdiff.scripts.svnforeign import query_unregistered_svn_files


def review_file(sobj, opts):
    """
    Check the given status object and if necessary, spawn xxdiff on it.

    Return a pair of ((file type description, action) waiter-object).
    """
    msg = ('normal', 'display')
    dopts = []
    merged = sobj.filename
    try:
        # Ignore unmodified files if there are any.
        if sobj.status in (' ', '?'):
            msg = ('unmodified', 'ignored')
            return msg, None

        # Diff modified files
        if sobj.status in ('M', 'C'):
            tmpf = subversion.cat_revision_temp(sobj.filename, 'BASE')
            left, right = tmpf.name, sobj.filename

            dopts.extend(title_opts('%s (BASE)' % sobj.filename))

        # Diff added files
        elif sobj.status == 'A':
            # Check if it is a directory.
            if not isfile(sobj.filename):
                msg = ('directory', 'skip')
                return msg, None

            if sobj.withhist == '+':
                # Get the source filename from the history.
                info = subversion.getinfo(sobj.filename)
                from_url, from_rev = [info.get('Copied From %s' % x, None)
                                      for x in ['URL', 'Rev']]

                tmpf = subversion.cat_revision_temp(sobj.filename, 'BASE')
                dopts.extend(title_opts('%s (%s)' % (from_url, from_rev)))
            else:
                tmpf = tempfile.NamedTemporaryFile(mode='w', prefix=tmpprefix)
                dopts.extend(title_opts('(NON-EXISTING)'))

            left, right = tmpf.name, sobj.filename

        # Diff deleted files
        elif sobj.status == 'D':
            tmpf = subversion.cat_revision_temp(sobj.filename, 'BASE')
            tmpf_empty = tempfile.NamedTemporaryFile(mode='w', prefix=tmpprefix)

            dopts.extend(title_opts('%s (BASE)' % sobj.filename,
                                    '(DELETED)'))

            left, right = tmpf.name, tmpf_empty.name

        # We don't know what to do with the rest yet.
        else:
            msg = ('unknown', 'ignored')
            print(("Error: Action for status '%s' on file '%s' "
                   "is not implemented yet") % (sobj.status, sobj.filename),
                  file=sys.stderr)
            return msg, None
    finally:
        pass

    # Check for non-text files.
    if not xxdiff.utils.istextfile(left) or not xxdiff.utils.istextfile(right):
        return ('non-text', 'skip'), None

    # Run xxdiff on the files.
    if opts.review:
        dopts.extend(['--decision'])
        opts.no_confirm = False
        opts.dry_run = False
        decision = xxdiff.condrepl.cond_replace(left, right, opts, sys.stdout,
                                                False, replfn=right)
        waiter = None
    else:
        waiter = xxdiff.invoke.xxdiff_display(opts, left, right, nowait=1, *dopts)

    return msg, waiter

def parse_options():
    """
    Parse the options.
    """
    import optparse
    parser = optparse.OptionParser(__doc__.strip())

    parser.add_option('-r', '--review', '--replace', action='store_true',
                      help="Review/replace mode: diff the file and replace"
                      "the target file by the output of the diff. Runs in "
                      "decision mode. This is useful in order to remove "
                      "selected changes before committing in.")

    parser.add_option('-c', '--commit', '--commit-with-comments',
                      action='store_true',
                      help="Spawns an editor window for adding comments and "
                      "starts the review, then commits the files with the "
                      "given comment.  NOTE: you editor is taken from the "
                      "environment variables XXDIFF_EDITOR, SVN_EDITOR, and "
                      "then EDITOR.  This program *MUST* open a new window "
                      "(for example, you cannot use just 'vi' but rather "
                      "something like: xterm -e 'vi %s').")

    o = parser.add_option('-C', '--comments-file', action='store',
                          default=None,
                          help="Specified the comments file to use.  This file "
                          "is left behind and its contents reused if it is "
                          "specified.")

    parser.add_option('-f', '--foreign', '--consider-foreign-files',
                      action='store_true',
                      help="Before starting the review/diffs, check all the "
                      "unregistered files and ask the user one by one about "
                      "what to do with them.")

    for mod in xxdiff.invoke, xxdiff.history:
        mod.options_graft(parser)
    xxdiff.backup.options_graft(parser,
                                "These options affect automatic backup of "
                                "deleted files, if enabled.")

    xxdiff.scripts.install_autocomplete(parser)

    opts, args = parser.parse_args()

    for mod in xxdiff.backup, xxdiff.invoke, xxdiff.history:
        mod.options_validate(opts, parser)

    if opts.comments_file and not opts.commit:
        print("(Option '%s' ignored.) " % o.dest, file=sys.stderr)

    # Add a verbose option for svnforeign to use.
    opts.verbose = True

    return opts, args


def svndiff_main():
    """
    Main program for svn-diff script.
    """
    opts, args = parse_options()

    # Compute the location of the resilient directory for the comments and
    # history files (and maybe more stuff later on).
    resildir = xxdiff.resilient.resilient_for_paths(args)
    hist = xxdiff.history.History(opts, resildir)

    # Compute a list of files to ignore (e.g. if the comments or history file is
    # located in the checkout, we want to ignore them and then associated
    # swap/temp files).
    ignofiles = []
    if opts.commit and opts.comments_file:
        comfn = abspath(opts.comments_file)
        ignofiles.extend([comfn, '%s.swp' % comfn])
    if opts.history and opts.history_file:
        histfn = abspath(opts.history_file)
        ignofiles.extend([histfn, '%s.swp' % histfn])

    if opts.foreign:
        # Consider the unregistered files.
        if query_unregistered_svn_files(
            args, opts, sys.stdout, ignore=ignofiles) is not True:
            # The user has quit, don't continue.
            sys.exit(0)
        print()
        print()

    # Get the status of the working copy.
    statii = subversion.status(args)

    # Ignore the comments file from the svn status output.
    statii = [s for s in statii if abspath(s.filename) not in ignofiles]

    if not statii:
        print('(Nothing to do, exiting.)')
        hist.delete()
        return

    # First print out the status to a string.
    renstatus = os.linesep.join(x.parsed_line for x in statii)

    if opts.commit:
        # File to delete after a successful commit.
        if opts.comments_file:
            comfn = abspath(opts.comments_file)
        else:
            # Select a comments file and make sure that it exists.
            comfn = join(resildir, 'comments')

            makedirs(resildir, False)
            if not exists(comfn):
                open(comfn, 'w')

    # Spawn an editor if requested before starting the review.
    if opts.commit:
        m = {'date': datetime.datetime.now()}
        comments = None # renstatus
        edit_waiter = xxdiff.editor.spawn_editor(comments, filename=comfn)

    # First print out the status to the user.
    print('Status Of Files To Be Diffed')
    print('----------------------------')
    print(renstatus)

    # Then we start printing each file and the associated decision.
    msgfmt = '  %-10s | %-10s | %s'
    print()
    print(msgfmt % ('Type', 'Action', 'Status'))
    print(msgfmt % ('-'*10, '-'*10, '-'*40))

    # Main loop for graphical diffs, over each of the files reported by status.
    for s in statii:
        kind, action = 'unknown', 'exception' # Initialize for in case of an
                                              # exception.
        try:
            # Skip directories.
            if isdir(s.filename):
                kind, action = 'directory', 'skip'
                continue
            elif islink(s.filename):
                kind, action = 'symlink', 'skip'
                continue

            # Compute unique string for history recorder.  We use the size, last
            # modification time, and status info to hash on this.
            if exists(s.filename):
                fstat = os.lstat(s.filename)
                sz, mtime = fstat.st_size, fstat.st_mtime
            else:
                # Deal with files that have been deleted.
                sz, mtime = 0, 0
            histitem = ' '.join((str(sz), str(mtime), s.parsed_line))

            # If the file has already been reviewed in the history, skip it.
            if hist.check(histitem):
                kind, action = 'seen', 'skip'
                continue

            # Review the file.
            (kind, action), waiter = review_file(s, opts)
        finally:
            print(msgfmt % (kind, action, s.parsed_line))

        if waiter is not None:
            waiter()

        # We have succesfully finished viewing the file, add it to the history.
        hist.append(histitem)

    # Commit the files if requested.
    if opts.commit:
        print("\nWaiting for editor '%s' to complete..." % edit_waiter.command, end='')
        sys.stdout.flush()
        comments = edit_waiter()
        print('Done.\n')
        print('Recorded Merge Comments: ', end='')
        if comments == '':
            print('(None)')
            comments = None
        else:
            print()
            print('-' * 70)
            print(comments)
            print('-' * 70)

        subversion.commit(args, comments=comments)

        # Delete temporary comments flie if we created it.
        if not opts.comments_file:
            xxdiff.resilient.resilient_remove(comfn)

    # The entire list of files has been reviewed (and possibly committed), clear
    # the history file.
    if opts.history:
        print()
        print('(Review complete, history cleared)')
        hist.delete()


def main():
    xxdiff.scripts.interruptible_main(svndiff_main)

if __name__ == '__main__':
    main()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""xx-cvs-diff [<options>] [<file> <file> ...]

This simple script invokes 'cvs diff' with the given file arguments, then splits
the output patch for individual files, applies the reverse patches to temporary
files and for each file it then spawns an xxdiff to preview each modified file
separately.  This allows you to preview the current changes that are made in a
cvs checkout.

Optionally, you can decide to modify or revert those changes and they are
committed file by file by the script.  In that case, the spawned xxdiff asks for
a decision by the user, then the script takes the following actions upon the
answer:

- ACCEPT: keep the local changes and and commit the file
- MERGED: merge by selecting diff hunmks and commit the merged file
- REJECT: throw away the local changes (do not commit)

For more generic behaviour about merging patches graphically, see also
xx-patch.  The current script is really about committing "some" cvs changes.
"""



__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4', 'cvs', 'diffutils']


# stdlib imports.
import sys, os, shutil
from subprocess import Popen, PIPE, STDOUT
from tempfile import NamedTemporaryFile

# xxdiff imports.
import xxdiff.scripts
import xxdiff.patches
import xxdiff.invoke
import xxdiff.backup
import xxdiff.scm.cvs
from xxdiff.scripts import tmpprefix


def parse_options():
    """
    Parse the options.
    """
    import optparse
    parser = optparse.OptionParser(__doc__.strip())

    parser.add_option('-c', '--commit', action='store_true',
                      help="Ask for confirmation and commit accepted changes.")

    xxdiff.scripts.install_autocomplete(parser)

    opts, args = parser.parse_args()
    return opts, args


def cvsdiff_main():
    """
    Main program for cvs-diff script.
    """
    opts, args = parse_options()

    # run cvs diff and read its output.

    cmd = 'cvs diff -u ' + ' '.join(['"%s"' % x for x in args])
    p = Popen(cmd, shell=True, stdout=PIPE, stderr=STDOUT, text=True)
    stdout, stderr = p.communicate()
    if p.returncode != 0 and not stdout:
        raise SystemExit("Error: running cvs command (%s): %s" %
                         (p.returncode, cmd))
    chunks = xxdiff.patches.splitpatch(stdout)

    logs = sys.stdout

    #
    # For each subpatch, apply it individually
    #
    for filename, patch in chunks:
        # print patch contents for this file.
        print('*' * 40)
        print(patch)
        print('*' * 40)

        # Feed diffs to patch, patch will do its deed and save the output to a
        # temporary file.
        tmpf = NamedTemporaryFile(prefix=tmpprefix)
        cin, cout = os.popen2(
            'patch --reverse --output "%s"' % tmpf.name, 'rw')
        cin.write('Index: %s\n' % filename)
        cin.write(patch)
        # avoid "patch unexpectedly ends in middle of line" warning.
        if patch[-1] != '\n':
            cin.write('\n')
        cin.close()

        # read output from patch.
        print(cout.read())

        if not opts.commit:
            # simply invoke xxdiff on the files.
            xxdiff.invoke.xxdiff_display(opts, tmpf.name, filename)
        else:
            decision, mergedf, retcode = xxdiff.invoke.xxdiff_decision(
                opts, '--title2', 'NEW FILE', tmpf.name, filename)

            # print output of xxdiff command.
            print(decision)

            # ACCEPT: keep the changes and and commit the file
            if decision == 'ACCEPT':
                xxdiff.scm.cvs.commit(filename)

            # MERGED: merge by selecting diff hunmks and commit the merged file
            elif decision == 'MERGED':
                # Save a backup, in case.
                xxdiff.backup.backup_file(filename, opts, logs)
                shutil.copyfile(mergedf.name, filename)
                xxdiff.scm.cvs.commit(filename)

            # REJECT: throw away the local changes (do not commit)
            elif decision == 'REJECT':
                xxdiff.backup.backup_file(filename, opts, logs)
                shutil.copyfile(tmpf.name, filename)

            # NODECISION: do nothing to the local file
            elif decision == 'NODECISION':
                pass


main = cvsdiff_main

if __name__ == '__main__':
    main()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions to backup files before overwriting them.
"""

__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
import sys, os, shutil, optparse, tempfile
from os.path import *

# xxdiff imports.
from xxdiff.scripts import tmpprefix


# Different kinds of backups.
backup_choices = ['along', 'other', 'none']

def options_graft(parser, msg=None, deftype='other'):
    """
    Graft options on given parser for automatic file backups.
    """
    if msg is None:
        msg = "These options affect automatic backup of overwritten files."

    group = optparse.OptionGroup(parser, "File backup options", msg)

    group.add_option('-b', '--backup-type', action='store',
                     type='choice', choices=backup_choices,
                     default=deftype, metavar="CHOICE",
                     help=("Selects the backup type from: %s (default: %s)" %
                           (', '.join(backup_choices), deftype)))

    group.add_option('--backup-dir', action='store',
                     help="Specify backup directory for type 'other'")

    group.add_option('--backup-prefix', action='store', default=None,
                     help=optparse.SUPPRESS_HELP)

    parser.add_option_group(group)

    return group


def options_validate(opts, parser, logs=None):
    """
    Validate backup options.
    """
    if opts.backup_type != 'other' and opts.backup_dir:
        parser.error(
            "option backup-dir is only valid for backups of type 'other'.")


def backup_file(fn, opts, logs=None):
    """
    Compute backup filename and copy backup file.

    Arguments:
    - 'fn': filename to backup -> string
    - 'opts': program options -> Options instance
    """
    if hasattr(logs, 'write'):
        logs = logs.write

    # Compute destination backup filename
    if opts.backup_type == 'along':
        # Search for a non-existing backup filename alongside the original
        # filename
        fmt = '%s.bak.%%d' % fn
        ii = 1
        while 1:
            backupfn = fmt % ii
            if not exists(backupfn):
                break
            ii += 1

    elif opts.backup_type == 'other':
        # Always backup the files rooted to the very root, because there may be
        # multiple root directories during the selection process.  Besides, it
        # does not matter all that much, the backup files are not really meant
        # to be accessed unless something goes really wrong.
        relfn = abspath(fn)

        # Remove the leading slash (this may f*ck up under Windoze).
        relfn = relfn[1:]

        # If the backup directory has not been set, create a temporary one
        if opts.backup_dir is None:
            if opts.backup_prefix:
                pfx = opts.backup_prefix + '.'
            else:
                pfx = ''
            opts.backup_dir = tempfile.mkdtemp(prefix=tmpprefix + pfx)
        # Otherwise create the specified directory if it does not exist
        elif not exists(opts.backup_dir):
            os.makedirs(opts.backup_dir)

        backupfn = join(opts.backup_dir, relfn)

        if opts.verbose >= 0 and exists(backupfn) and logs:
            logs("(Warning: Overwriting existing file in backup '%s')\n" %
                 backupfn)

    else: # opts.backup_type == 'none'
        backupfn = None

    if backupfn:
        # Perform the backup
        if logs and opts.verbose >= 3:
            logs('Backup: %s\n' % backupfn)

        # Make sure that the destination directory exists
        ddn = dirname(backupfn)
        if ddn and not exists(ddn):
            os.makedirs(ddn)

        # Copy the original to the backup directory
        if isfile(fn) or islink(fn):
            shutil.copy2(fn, backupfn)
        elif isdir(fn):
            shutil.copytree(fn, backupfn, True)
        else:
            raise RuntimeError("Internal error: cannot backup '%s'. " % fn,
                               "Invalid file type.")

    return backupfn


def print_reminder(opts):
    """
    Print a reminder of the location of the backup files.
    """
    # Print reminder of location of backup files at the end (for convenience).
    if opts.backup_dir:
        print()
        print(("Backup files in: '%s'" % opts.backup_dir))
        print()


def test():
    """
    Test stub.
    """
    parser = optparse.OptionParser(__doc__.strip())
    options_graft(parser)
    opts, args = parser.parse_args()

    options_validate(opts, parser, logs=sys.stdout)

    print(('Dir:', opts.backup_dir))
    print(('Type:', opts.backup_type))

    backup_file('/home/blais/.emacs', opts, sys.stdout)



if __name__ == '__main__':
    test()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""xx-svn-resolve [<options>] <file> <file> [<file>]

Script that invokes xxdiff with a 3-way merge view for all files in a Subversion
checkout whose conflicts need to be resolved.  Depending on the exit status,
xxdiff also calls 'svn resolve' on the files.
"""



__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4', 'Subversion']

# stdlib imports.
import sys, os, re
from os.path import *

# xxdiff imports.
import xxdiff.scripts
import xxdiff.backup
import xxdiff.condrepl
from xxdiff.scm import subversion
from xxdiff.scripts.encrypted import diff_encrypted


def parse_options():
    """
    Parse the options.
    """
    import optparse
    parser = optparse.OptionParser(__doc__.strip())

    xxmodules = (xxdiff.backup,)
    for mod in xxmodules:
        mod.options_graft(parser)

    parser.add_option('-n', '--dry-run', action='store_true',
                      help="Print the commands that would be executed " +
                      "but don't really run them.")

    parser.add_option('-R', '--no-resolve', '--dont-resolve',
                      action='store_true',
                      help="Do not resolve the conflicts even after a merge "
                      "decision has been made.")

    # Some of the GPG options duplicated here.
    parser.add_option('-g', '--gpg', default="gpg",
                      help="Specify path to gpg program to use.")
    parser.add_option('-A', '--dont-armor', action='store_true',
                      help="Create output file in binary format.")
    parser.add_option('-r', '--recipient', action='store',
                      help="Encrypt for user id name.")

    xxdiff.scripts.install_autocomplete(parser)
    opts, args = parser.parse_args()

    # Force to always perform a verbose diff on output.
    opts.verbose = 2

    for mod in xxmodules[1:]:
        mod.options_validate(opts, parser)

    return opts, args


def select_conflicts(statii):
    """
    A generator that selects only the conflictual files.
    """
    for s in statii:
        if s.status == 'C':
            yield s

def svnresolve_main():
    """
    Main program for svn-resolve script.
    """
    opts, args = parse_options()

    # Get the status of the working copy.
    statii = subversion.status(args)

    # First print out the list/status of the conflicting files to the user.
    for s in select_conflicts(statii):
        print(s.parsed_line)
    print()

    logs = sys.stdout

    # For each of the files reported by status
    for s in select_conflicts(statii):

        # Get the three files before the merge conflicts.
        info = subversion.getinfo(s.filename)
        dn = dirname(s.filename)
        ancestor, mine, yours = [join(dn, info[x]) for x in
                                 ('Conflict Previous Base File',
                                  'Conflict Previous Working File',
                                  'Conflict Current Base File')]

        # Spawn xxdiff in decision mode on the three files. We dispatch to the
        # encrypted version if necessary.
        if re.match('.*\.asc', s.filename):
            tmine = open(mine).read()
            tancestor = open(ancestor).read()
            tyours = open(yours).read()
            decision = diff_encrypted([tmine, tancestor, tyours], opts, outmerged=s.filename)
        else:
            decision = xxdiff.condrepl.cond_resolve(
                mine, ancestor, yours, s.filename, opts, logs, extra=('--merge',))

        # Backup all the other files that will get when the file gets resolved,
        # whether by this script or later by the user by hand.
        if decision in xxdiff.condrepl.proper_decisions:
            xxdiff.backup.backup_file(mine, opts, logs)
            xxdiff.backup.backup_file(ancestor, opts, logs)
            xxdiff.backup.backup_file(yours, opts, logs)
        else:
            # If no proper decision has been made, do not resolve.
            continue

        # Resolve the conflict with Subversion, if requested.
        if not opts.no_resolve:
            subversion.resolve(s.filename)
            print('(Resolved.)')

    xxdiff.backup.print_reminder(opts)


def main():
    xxdiff.scripts.interruptible_main(svnresolve_main)

if __name__ == '__main__':
    main()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""xx-sql-schemas [<options>] <db1> <db2>

Where <dbX> is of the form:   [user@]<dbname>[/schema]

Script that dumps the schemas for two PostgreSQL databases, re-orders the
contents so that they diff nicely, and spawns xxdiff on the results.

Note: for now this script only supports the PostgreSQL database.  If you want to
add support for another database, just add an option and an appropriate
xxdiff.db.* module rather than writing a dedicated script.  (I'm very flexible
about adding new features to this script.)
"""



__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4', 'PostgreSQL']

# stdlib imports.
from os.path import *
import logging, shutil

# xxdiff imports.
import xxdiff.scripts
import xxdiff.invoke
import xxdiff.mapcompare
from xxdiff.db import postgresql


def parse_options():
    """
    Parse the options.
    """
    import optparse
    parser = optparse.OptionParser(__doc__.strip())

    xxdiff.mapcompare.options_graft(parser)
    postgresql.options_graft(parser)
    xxdiff.invoke.options_graft(parser)

    parser.add_option('-t', '--text-only', action='store_true',
                      help="Render as text files to disk instead of graphical diff.")

    opts, args = parser.parse_args()

    xxdiff.mapcompare.options_validate(opts, parser)
    postgresql.options_validate(opts, parser)
    xxdiff.invoke.options_validate(opts, parser)

    if len(args) != 2:
        parser.error("You need to specify two databases/schemas to compare.")

    dbspec1, dbspec2 = args

    return (opts,
            postgresql.parse_dbspec(dbspec1, parser, opts),
            postgresql.parse_dbspec(dbspec2, parser, opts))

def sqlcompare_main():
    """
    Main program for schema comparison script.
    """
    opts, db1, db2 = parse_options()

    # Dump the database schemas.
    for db in db1, db2:
        dump = postgresql.dump_schema(db.user, db.dbname, db.schema, opts)

        # Parse the dumps, produce an adequate map of its contained objects.
        dump_entries = postgresql.parse_dump(dump)
        db.objmap = dict(((name, type), contents)
                         for name, type, contents in dump_entries)

    # List all objects that were found.
    for key in sorted(set(list(db1.objmap.keys()) + list(db2.objmap.keys()))):
        print(key)

    # Render comparable files.
    files = xxdiff.mapcompare.render_diffable_maps(opts, db1.objmap, db2.objmap)

    # Set displayed titles.
    if opts.text_only:
        for db, f in zip([db1, db2], files):
            targetfn = '%s.schema' % db.dbname
            logging.info("Writing out %s" % targetfn)
            shutil.copyfile(f.name, targetfn)
    else:
        titles = xxdiff.invoke.title_opts(
            *['%s (SCHEMA)' % db.dbspec for db in (db1, db2)])

        # Invoke xxdiff .
        xxargs = titles + [x.name for x in files]
        xxdiff.invoke.xxdiff_display(opts, *xxargs)


def main():
    xxdiff.scripts.interruptible_main(sqlcompare_main)

if __name__ == '__main__':
    main()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions for Subversion.

Note: we do not use the official Python bindings because we want to evolve
independently and not break when Subversion changes.  This only makes sense
inasmuch as we do not do much with Subversion.  Should we get do more involved
stuff with it, we should switch this code to use the Subversion Python bindings.
"""



__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
import os, tempfile, re
from subprocess import Popen, call, PIPE
from os.path import abspath, isabs

# xxdiff imports.
from xxdiff.scripts import script_name, tmpprefix


def is_checked_out(filename):
    """
    There is no real global mutual editing policy in Subversion.  The restricted
    locking mechanism that is available is not meant to be automated (on the
    contrary).
    """
    return True


def checkout(filename):
    """
    Checkout the given filename from Subversion.
    """
    return


def commit(filenames, comments=None):
    """
    Commit the given filename into CVS.
    """
    if comments is None:
        comments = 'Committed by %s' % script_name

    if isinstance(filenames, str):
        filenames = [filenames]
    else:
        assert isinstance(filenames, (tuple, list))

    # Note: if the comments are ever so huge, we consider using a file with the
    # -f option to 'svn commit'.  Wondering if there may be any important issues
    # with passing large objects through the environment.  This is certainly not
    # the usual thing to do.
    if len(comments) < 512:
        call(['svn', 'commit', '--message', comments] + filenames)
    else:
        tmpf = tempfile.NamedTemporaryFile(mode='w', prefix=tmpprefix)
        tmpf.write(comments)
        tmpf.flush()
        call(['svn', 'commit', '--file', tmpf.name] + filenames)


def resolve(filename):
    """
    Commit the given filename into CVS.
    """
    p = Popen(['svn', 'resolved', filename], stdout=PIPE, stderr=PIPE, text=True)
    stdout, stderr = p.communicate()
    if p.returncode != 0:
        raise SystemExit("Error: Resolving file '%s':\n%s\n" %
                         (filename, stderr))

class SvnStatus(object):
    """
    Status result for a file in a subversion repository.
    """

def status(rootdirs):
    """
    Obtains the status from all the given root directories.
    A list of status objects is returned.
    """
    if isinstance(rootdirs, str):
        rootdirs = [rootdirs]
    else:
        assert isinstance(rootdirs, (list, tuple))
    if len(rootdirs) == 0:
        rootdirs.append(os.getcwd())

    # Note: since we specify abspaths we should get abspaths on the output
    p = Popen(['svn', 'status'] + list(map(abspath, rootdirs)), stdout=PIPE, text=True)
    stdout, stderr = p.communicate()

    statii = []
    for line in stdout.splitlines():
        if not line or re.match('^Performing', line):
            continue

        status = SvnStatus()
        status.parsed_line = line

        # The first six columns in the output are each one character wide:
        lc = iter(line)

        #   First column: Says if item was added, deleted, or otherwise
        #   changed
        #     ' ' no modifications
        #     'A' Added
        #     'C' Conflicted
        #     'D' Deleted
        #     'G' Merged
        #     'I' Ignored
        #     'M' Modified
        #     'R' Replaced
        #     'X' item is unversioned, but is used by an externals
        #         definition
        #     '?' item is not under version control
        #     '!' item is missing (removed by non-svn command) or incomplete
        #     '~' versioned item obstructed by some item of a different kind
        status.status = next(lc)

        #   Second column: Modifications of a file's or directory's
        #   properties
        #     ' ' no modifications
        #     'C' Conflicted
        #     'M' Modified
        status.modprops = next(lc)

        #   Third column: Whether the working copy directory is locked
        #     ' ' not locked
        #     'L' locked
        status.dirlocked = next(lc)

        #   Fourth column: Scheduled commit will contain
        #   addition-with-history
        #     ' ' no history scheduled with commit
        #     '+' history scheduled with commit
        status.withhist = next(lc)

        #   Fifth column: Whether the item is switched relative to its
        #   parent
        #     ' ' normal
        #     'S' switched
        status.switched = next(lc)

        #   Sixth column: Repository lock token
        #     (without -u)
        #     ' ' no lock token
        #     'K' lock token present
        #     (with -u)
        #     ' ' not locked in repository, no lock token
        #     'K' locked in repository, lock toKen present
        #     'O' locked in repository, lock token in some Other working
        #          copy
        #     'T' locked in repository, lock token present but sTolen
        #     'B' not locked in repository, lock token present but Broken
        status.locktoken = next(lc)

        #   Seventh column: Repository lock token
        #     'C'   (conflict on merged files)
        #     '>'   (conflict comments)
        status.mconflict = next(lc)
        if status.mconflict == '>':
            continue

        # There is a space and the rest is a filename:
        status.filename = line[7:].strip()
        assert isabs(status.filename), (
            f"Filename {status.filename} is not a path. Check output of svn-foreign.")

        statii.append(status)


    # Note: we could parse this information, but this requires server access:
    # -----------------------------------------------------------------------
    #   The out-of-date information appears in the eighth column (with -u):
    #       '*' a newer revision exists on the server
    #       ' ' the working copy is up to date
    #
    #   Remaining fields are variable width and delimited by spaces:
    #     The working revision (with -u or -v)
    #     The last committed revision and last committed author (with -v)
    #     The working copy path is always the final field, so it can
    #       include spaces.

    return statii





def getinfo(filename):
    """
    Return the fields of 'svn info', in a dictionary.
    """
    p = Popen(['svn', 'info', filename], stdout=PIPE, stderr=PIPE, text=True)
    stdout, stderr = p.communicate()
    if p.returncode != 0:
        raise SystemExit("Error: Cannot get history for: %s" % stderr)

    colon_split_re = re.compile(': ')
    d = dict([colon_split_re.split(x, 1) for x in stdout.strip().splitlines()])

    return d


def cat_revision_temp(filename, revision):
    """
    Fetches a specific revision of a file and place it in a temporary file,
    which is returned.
    """
    # Fetch the local ancestor to a temporary file.
    tmpf = tempfile.NamedTemporaryFile(mode='w', prefix=tmpprefix)

    p = Popen(['svn', 'cat', '--revision', revision, filename],
              stdout=tmpf, text=True)
    stdout, stderr = p.communicate()
    if p.returncode != 0:
        raise SystemExit("Error: Cannot fetch BASE revision: %s" % stderr)

    return tmpf





def test():
    """
    Simple interactive test.
    """
    import optparse
    parser = optparse.OptionParser(__doc__.strip())
    opts, args = parser.parse_args()

    for s in status(args[0]):
        print(s.filename, ':', s.status)


if __name__ == '__main__':
    test()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions to manipulate patches.
"""

__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
import re


def splitpatch(text):
    """
    Split output in chunks starting with ^Index.  Returns a list of pairs
    (tuples), each with (filename, patch) contents.
    """

    ## splitre = re.compile('^Index: (.*)$', re.M)
    splitre = re.compile('^Index: (.*)$', re.M)
    chunks = []
    curbeg, curfn = None, None
    for mo in splitre.finditer(text):
        if curbeg != None:
            assert curfn
            chunks.append( (curfn, text[curbeg:mo.start()]) )
        curbeg = mo.start()
        curfn = mo.group(1)
    if curbeg != None:
        chunks.append( (curfn, text[curbeg:]) )

    return chunks





# This file is part of the xxdiff package.  See xxdiff for license and details.

"""xx-cond-replace [<options>] <orig-file> <modified-file>

Useful script to conditionally replace a original by a generated file.

Basically, you invoke some script or pipeline of scripts to perform
modifications on an original text file and store the results in a temporary file
(the modified-file), and then invoke this script to conditionally overwrite the
contents of the orig-file with those of the modified file.

This works like copy (``cp``), except that xxdiff is shown to ask the user
whether to copy the file or not.  This is useful to run in a loop, for example::

   for i in `*/*.xml` ; do
      cat $i | ...pipeline of cmds that modifies the file... > /tmp/out.xml
      # copy only if the user accepts or merges.
      xx-cond-replace /tmp/out.xml $i
   done

**IMPORTANT**: Notice well that the original file which will be overwritten is
on the RIGHT side. The syntax is like that of the UNIX cp command.

Exit Status
-----------

This program exits with status 0 if the copy operation was accepted or merged,
with 1 otherwise.

Notes
-----

- the script automatically creates backup files.

- the script automatically generates a detailed log of its actions and
  a text summary of all the differences beteween the original and new files.

- the script can optionally checkout the file before performing the replacement.

"""

__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4']


# stdlib imports.
import sys, os
from os.path import *
import tempfile

# xxdiff imports.
import xxdiff.scripts
import xxdiff.backup
import xxdiff.checkout
import xxdiff.invoke
import xxdiff.condrepl
from xxdiff.scripts import tmpprefix


def parse_options():
    """
    Parse the options.
    """
    import optparse

    parser = optparse.OptionParser(__doc__.strip())

    xxmodules = xxdiff.checkout, xxdiff.invoke, xxdiff.condrepl
    for mod in xxmodules:
        mod.options_graft(parser)
    xxdiff.backup.options_graft(parser, deftype='none')

    parser.add_option('-v', '--verbose', action='count',
                      default=0,
                      help="Verbose output.  By default nothing is output. "
                      "If you use this switch once, you will get the decision "
                      "code followed by the filename.  If you use it twice, "
                      "you will obtain a side-by-side diff of the changes.")

    parser.add_option('-d', '--diff', action='store_true',
                      help="Run a diff and log the differences on stdout.")

    parser.add_option('--delete', action='store_true',
                      help="Instead of copying the temporary file, move it "
                      "(delete it after copying).")

    parser.add_option('-D', '--exit-on-same', action='store_true',
                      help="Do not do anything if the files are the same.")

    xxdiff.scripts.install_autocomplete(parser)
    opts, args = parser.parse_args()

    for mod in xxmodules:
        mod.options_validate(opts, parser, logs=sys.stdout)

    if len(args) != 2:
        raise parser.error("you must specify exactly two files.")
    fromfn, tofn = args

    if fromfn == '-' and opts.delete:
        raise parser.error("no need to use --delete on file from stdin.")
    if tofn == '-':
        raise parser.error("cannot copy to stdin, this does not make sense.")

    return opts, fromfn, tofn


def condreplace_main():
    """
    Main program for cond-replace script.
    """
    opts, fromfn, tofn = parse_options()

    if fromfn == '-':
        intmpf = tempfile.NamedTemporaryFile(mode='w', prefix=tmpprefix)
        try:
            intmpf.write(sys.stdin.read())
            sys.stdin.close()
            intmpf.flush()
        except IOError as e:
            raise SystemExit(
                "Error: saving stdin to temporary file (%s)" % str(e))
        fromfn = intmpf.name


    # call xxdiff and perform the conditional replacement.
    decision = xxdiff.condrepl.cond_replace(tofn, fromfn, opts, sys.stdout,
                                            opts.exit_on_same)

    if opts.delete and not opts.dry_run:
        try:
            os.unlink(fromfn)
        except OSError as e:
            raise SystemExit("Error: deleting modified file (%s)" % str(e))

    # repeat message at the end for convenience.
    if opts.verbose > 0:
        xxdiff.backup.print_reminder(opts)

    # Compute return value.
    rval = 0
    if decision in ('REJECT', 'NODECISION'):
        rval = 1

    return rval


def main():
    xxdiff.scripts.interruptible_main(condreplace_main)

if __name__ == '__main__':
    main()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions to record and recall history of actions take on lists of files.
"""

__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
import os, optparse
from os.path import *

# xxdiff imports.
import xxdiff.resilient
from xxdiff.scripts import script_name
from xxdiff.utils import makedirs


def options_graft(parser):
    """
    Graft options on given parser for invocations of xxdiff.
    """
    group = optparse.OptionGroup(parser, "Options for action history",
                                 "These options affect the history of "
                                 "files/actions.")

    group.add_option('--no-history', action='store_false',
                     dest='history', default=True,
                     help="Do not record to the file history.")

    group.add_option('--clear-history', '--reset-history', action='store_true',
                     help="Clears the file history before starting.")

    group.add_option('--history-file', action='store', metavar='FILENAME',
                     help="Use the specified file to record history.")

    parser.add_option_group(group)

    return group


def options_validate(opts, parser, logs=None):
    """
    Validate xxdiff options.
    """
    if not opts.history and opts.history_file:
        parser.error("You cannot specify a history file and disable history.")


class History(object):
    """
    A class that contains the runtime state of the history of some multi-item
    process/actions.

    The history is saved one file per-line, but you can put whatever you like in
    that one line, for example, the filename, the status of the file, its last
    modification time, etc.  It is basically used as a hash.

    Note: eventually we might want to put more stuff in the history item.  We
    could then consider using a pickle instead.  I have found this unnecessary
    for the current uses of this class (i.e. recording the state of a merge
    preview).

    This is used to record a history of files previewed in a diff loop, so that
    if the viewer loop is interrupted it can be restarted without having to
    review all the past differences.
    """
    def __init__(self, opts, resilient_dir):
        self.history = []
        "Existing history file lines."

        self.histf = None
        """A file object opened in write mode so that we can write new entries
        there."""

        self.opts = opts
        "A reference to the options for use later."
        
        self._initialize(resilient_dir)

    def _initialize(self, resilient_dir):
        """
        Initialize the history module.
        """
        opts = self.opts
        
        # Calculate the location of the history file.
        if opts.history_file:
            histfn = opts.history_file
        else:
            histfn = join(resilient_dir, 'history')

        # Clear the history if requested.
        if opts.clear_history and exists(histfn):
            os.unlink(histfn)

        # Prepare and create a history file if one does not already exist.
        if not opts.history:
            return
        if not exists(histfn):
            makedirs(dirname(histfn), False)
            self.histf = open(histfn, 'w')

        # Otherwise read the history file contents.
        else:
            f = open(histfn, 'r')
            self.history.extend(x.strip() for x  in f.readlines() if x)
            f.close()
            self.histf = open(histfn, 'a')

    def __contains__(self, histitem):
        """
        Support 'in' syntax.
        """
        return self.check(histitem)

    def check(self, histitem):
        """
        Check if the given history item has already been seen.  Call this to
        check every item to be run on your process to be run and to skip it if
        it has already been seen.
        """
        # If history is disabled, check always returns false.
        if not self.opts.history:
            return False

        return histitem in self.history

    def append(self, histitem):
        """
        Append the given history item to the current history and save the file.
        Call this for every succesfully completed item of your process
        """
        # If history is disabled, we don't do anything on append.
        if not self.opts.history:
            return

        # Don't add itmes which are already present in the history.
        if histitem in self.history:
            return

        # Append the new item to the history.
        histitem = histitem.strip()
        self.history.append(histitem)
        
        # Update the history file and make sure that all changes are flushed to
        # disk so that if we get interrupted nothing is lost.
        self.histf.write(histitem)
        self.histf.write('\n')
        self.histf.flush()

    def delete(self):
        """
        Delete the history file.  (This also clears the history.)
        Call this when your process completed succesfully.
        """
        histfn = self.histf.name
        self.histf.close()
        xxdiff.resilient.resilient_remove(histfn)




# This file is part of the xxdiff package.  See xxdiff for license and details.

"""xx-match [<options>] <file> <file> <file> <file> ...

Wrapper script for xxdiff that implements an invocation syntax like this:

  xx-match dir1/*.c dir2/*.c

The way we implement the heuristic to match up files is simple: we match up the
basenames of the files.  If there are exactly two or three files in the list, we
invoke an xxdiff on these files.  The files are always invoked in the order that
they show up on the command-line.

Another way to achieving similar behaviour is to use a more generic file
matching script (see 'match-files' by the same author) along with xargs, e.g.::

  match-files dir1/*.c dir2/*.c | xargs -n2 xxdiff

Using xx-match has the advantage that it work with matches of 2 or 3 files,
and also spawns xxdiff on single files (with an empty side).

Note: many users do not seem to understand why such a script is necessary. One
must realize a few things: 1) globbing pattern expansion is done by the shell,
before invocation of xxdiff, thus xxdiff cannot be aware of the globbing
patterns (unless it itself would support the heuristic mentioned here, along
with the user quoting the globbing patterns to avoid shell expansion); 2) the
solution to this problem using expanded file lists is not trivial, the input
(and its ordering) can encompass a lot more possibilities than immediately
appears. Nevertheless, we built this functionality/heuristic in this script,
because it may be useful to some.

See 'match-files' script by the same author for a more generic implementation of
this file matching heuristic.
"""

__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4']


# stdlib imports.
import sys, os

# xxdiff imports.
import xxdiff.scripts
import xxdiff.invoke


def match_main():
    """
    Main program for match script.
    """
    # Get options
    opts, files = xxdiff.scripts.passthruopts(sys.argv)

    # Make sure that we display the commands.
    class Opts:
        xxdiff_verbose = True

    # Build map of basenames
    bnmap = {}
    for fn in files:
        dn, bn = os.path.split(fn)
        bnmap.setdefault(bn, []).append(fn)

    # Invoke xxdiff's on alphabetical order of the basenames
    bnkeys = list(bnmap.keys())
    bnkeys.sort()
    for bn in bnkeys:
        filenames = bnmap[bn]
        if len(filenames) == 2 or len(filenames) == 3:
            extra = []
        elif len(filenames) == 1:
            extra = ['--single']
        else:
            # ignore the files.
            continue

        xxdiff.invoke.xxdiff_display(Opts, *(extra + filenames))


def main():
    xxdiff.scripts.interruptible_main(match_main)

if __name__ == '__main__':
    main()




# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Resilient files.

This is used to manage a location where merge comments and preview history are
saved between sessions.

Support a per-user directory to hold resilient temporary files.  We want to be
able to recycle the files stored in there if the process/script has been
interrupted for some reason--this is quite common for source code, as the
reviewer often realizes he forgot to do something or other, this is a benefit of
reviewing diffs before committing--if the same directories/files are to be
checked-in.  We do not want to store the files in the CWD because that might
change, and it may later interfere with other tools, e.g. show up in 'svn
status' if left behind.  Therefore, we store the comments file in the user's
home directory under a hash computed from the absolute paths of the directories
given as arguments.
"""

__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
import os, hashlib
from os.path import *

# xxdiff imports.
from xxdiff.scripts import script_name


resilient_dir = join(os.environ['HOME'], '.%s' % script_name)

def resilient_for_paths(paths):
    """
    Compute a unique file or directory name for a set of paths (considered
    unique as absolute) for a specific process.
    """

    # Compute a unique hash from the given absolute paths.
    comhash = hashlib.md5()
    for p in paths:
        comhash.update(abspath(p))
    resdir = join(resilient_dir, comhash.hexdigest())
    return resdir

def resilient_remove(fn):
    """
    Removes a resilient file, and then checks if the resilient file's parent
    directory can be removed as well.
    """
    os.unlink(fn)
    parent = dirname(fn)
    if not os.listdir(parent):
        os.rmdir(parent)


